@@ -10,7 +10,7 @@ import math
 import pprint
 import logging
 import datetime
-from typing import Tuple, List, Dict, Any, Union
+from typing import Tuple, List, Dict, Iterable, Any, Union
 from pygments import highlight, lexers, formatters
 from collections import OrderedDict
 from collections.abc import Sized
@@ -26,10 +26,12 @@ from stefutil.primitive import is_float
 
 
 __all__ = [
-    'fmt_num', 'fmt_sizeof', 'fmt_delta', 'sec2mmss', 'round_up_1digit', 'nth_sig_digit', 'now',
+    'fmt_num', 'fmt_sizeof', 'fmt_delta', 'sec2mmss', 'round_up_1digit', 'nth_sig_digit',
     'MyIceCreamDebugger', 'mic',
-    'log', 'log_s', 'logi', 'log_dict', 'log_dict_nc', 'log_dict_id', 'log_dict_pg', 'log_dict_p',
+    'PrettyLogger', 'pl',
     'hex2rgb', 'MyTheme', 'MyFormatter', 'get_logger',
+    'CheckArg', 'ca',
+    'now',
     'MlPrettier', 'MyProgressCallback'
 ]
 
@@ -38,6 +40,7 @@ pd.set_option('expand_frame_repr', False)
 pd.set_option('display.precision', 2)
 pd.set_option('max_colwidth', 40)
 pd.set_option('display.max_columns', None)
+pd.set_option('display.min_rows', 16)
 
 
 def fmt_num(num: Union[float, int], suffix: str = '') -> str:
@@ -93,17 +96,6 @@ def nth_sig_digit(flt: float, n: int = 1) -> float:
     return float('{:.{p}g}'.format(flt, p=n))
 
 
-def now(as_str=True, for_path=False) -> Union[datetime.datetime, str]:
-    """
-    # Considering file output path
-    :param as_str: If true, returns string; otherwise, returns datetime object
-    :param for_path: If true, the string returned is formatted as intended for file system path
-    """
-    d = datetime.datetime.now()
-    fmt = '%Y-%m-%d_%H-%M-%S' if for_path else '%Y-%m-%d %H:%M:%S'
-    return d.strftime(fmt) if as_str else d
-
-
 class MyIceCreamDebugger(IceCreamDebugger):
     def __init__(self, output_width: int = 120, **kwargs):
         self._output_width = output_width
@@ -126,107 +118,151 @@ class MyIceCreamDebugger(IceCreamDebugger):
 mic = MyIceCreamDebugger()
 
 
-def log(s, c: str = 'log', c_time='green', as_str=False, bold: bool = False, pad: int = None):
+class PrettyLogger:
     """
-    Prints `s` to console with color `c`
+    My logging w/ color & formatting, and a lot of syntactic sugar
     """
-    if not hasattr(log, 'reset'):
-        log.reset = colorama.Fore.RESET + colorama.Back.RESET + colorama.Style.RESET_ALL
-    if not hasattr(log, 'd'):
-        log.d = dict(
-            log='',
-            warn=colorama.Fore.YELLOW,
-            error=colorama.Fore.RED,
-            err=colorama.Fore.RED,
-            success=colorama.Fore.GREEN,
-            suc=colorama.Fore.GREEN,
-            info=colorama.Fore.BLUE,
-            i=colorama.Fore.BLUE,
-            w=colorama.Fore.RED,
+    reset = colorama.Fore.RESET + colorama.Back.RESET + colorama.Style.RESET_ALL
+    key2c = dict(
+        log='',
+        warn=colorama.Fore.YELLOW,
+        error=colorama.Fore.RED,
+        err=colorama.Fore.RED,
+        success=colorama.Fore.GREEN,
+        suc=colorama.Fore.GREEN,
+        info=colorama.Fore.BLUE,
+        i=colorama.Fore.BLUE,
+        w=colorama.Fore.RED,
 
-            y=colorama.Fore.YELLOW,
-            yellow=colorama.Fore.YELLOW,
-            red=colorama.Fore.RED,
-            r=colorama.Fore.RED,
-            green=colorama.Fore.GREEN,
-            g=colorama.Fore.GREEN,
-            blue=colorama.Fore.BLUE,
-            b=colorama.Fore.BLUE,
+        y=colorama.Fore.YELLOW,
+        yellow=colorama.Fore.YELLOW,
+        red=colorama.Fore.RED,
+        r=colorama.Fore.RED,
+        green=colorama.Fore.GREEN,
+        g=colorama.Fore.GREEN,
+        blue=colorama.Fore.BLUE,
+        b=colorama.Fore.BLUE,
 
-            m=colorama.Fore.MAGENTA
-        )
-    if c in log.d:
-        c = log.d[c]
-    if bold:
-        c += colorama.Style.BRIGHT
-    if as_str:
-        return f'{c}{s:>{pad}}{log.reset}' if pad is not None else f'{c}{s}{log.reset}'
-    else:
-        print(f'{c}{log(now(), c=c_time, as_str=True)}| {s}{log.reset}')
+        m=colorama.Fore.MAGENTA
+    )
 
-
-def log_s(s, c, bold: bool = False):
-    return log(s, c=c, as_str=True, bold=bold)
-
-
-def logi(s):
-    """
-    Syntactic sugar for logging `info` as string
-    """
-    return log_s(s, c='i')
-
-
-def log_dict(d: Dict = None, with_color=True, pad_float: int = 5, sep=': ', **kwargs) -> str:
-    """
-    Syntactic sugar for logging dict with coloring for console output
-    """
-    def _log_val(v):
-        if isinstance(v, dict):
-            return log_dict(v, with_color=with_color)
+    @staticmethod
+    def log(s, c: str = 'log', c_time='green', as_str=False, bold: bool = False, pad: int = None):
+        """
+        Prints `s` to console with color `c`
+        """
+        need_reset = False
+        if c in PrettyLogger.key2c:
+            c = PrettyLogger.key2c[c]
+            need_reset = True
+        if bold:
+            c += colorama.Style.BRIGHT
+            need_reset = True
+        reset = PrettyLogger.reset if need_reset else ''
+        if as_str:
+            return f'{c}{s:>{pad}}{reset}' if pad is not None else f'{c}{s}{reset}'
         else:
-            if is_float(v):  # Pad only normal, expected floats, intended for metric logging
-                if is_float(v, no_int=True, no_sci=True):
-                    v = float(v)
-                    return log(v, c='i', as_str=True, pad=pad_float) if with_color else f'{v:>{pad_float}}'
-                else:
-                    return logi(v) if with_color else v
+            print(f'{c}{PrettyLogger.log(now(), c=c_time, as_str=True)}| {s}{reset}')
+
+    @staticmethod
+    def s(s, c: str = None, bold: bool = False, with_color: bool = True) -> str:
+        """
+        syntactic sugar for return string instead of print
+        """
+        c = c if with_color else ''  # keeping the same signature with logging specific types for `lognc`
+        return PrettyLogger.log(s, c=c, as_str=True, bold=bold)
+
+    @staticmethod
+    def i(s, **kwargs):
+        """
+        Syntactic sugar for logging `info` as string
+        """
+        if isinstance(s, dict):
+            return PrettyLogger._dict(s, **kwargs)
+        elif isinstance(s, list):
+            return PrettyLogger._list(s, **kwargs)
+        elif isinstance(s, tuple):
+            return PrettyLogger._tuple(s, **kwargs)
+        else:
+            kwargs_ = dict(c='i')
+            kwargs_.update(kwargs)
+            return PrettyLogger.s(s, **kwargs_)
+
+    @staticmethod
+    def pa(s):
+        assert isinstance(s, dict)
+        PrettyLogger.i(s, for_path=True, with_color=False)
+
+    @staticmethod
+    def nc(s, **kwargs):
+        """
+        Syntactic sugar for `i` w/o color
+        """
+        return PrettyLogger.i(s, with_color=False, **kwargs)
+
+    @staticmethod
+    def id(d: Dict) -> str:
+        """
+        Indented
+        """
+        return json.dumps(d, indent=4)
+
+    @staticmethod
+    def fmt(s) -> str:
+        """
+        colored by `pygments` & with indent
+        """
+        return highlight(PrettyLogger.id(s), lexers.JsonLexer(), formatters.TerminalFormatter())
+
+    @staticmethod
+    def _iter(it: Iterable, with_color=True, pref: str = '[', post: str = ']'):
+        if with_color:
+            pref, post = PrettyLogger.s(pref, c='m'), PrettyLogger.s(post, c='m')
+        lst = [PrettyLogger.i(e, with_color=with_color) for e in it]
+        return f'{pref}{", ".join(lst)}{post}'
+
+    @staticmethod
+    def _list(lst: List, with_color=True):
+        return PrettyLogger._iter(lst, with_color=with_color, pref='[', post=']')
+
+    @staticmethod
+    def _tuple(tpl: Tuple, with_color=True):
+        return PrettyLogger._iter(tpl, with_color=with_color, pref='(', post=')')
+
+    @staticmethod
+    def _dict(d: Dict = None, with_color=True, pad_float: int = 5, sep=': ', for_path: bool = False, **kwargs) -> str:
+        """
+        Syntactic sugar for logging dict with coloring for console output
+        """
+        def _log_val(v):
+            if isinstance(v, (dict, list, tuple)):
+                return PrettyLogger.i(v, with_color=with_color)
             else:
-                return logi(v) if with_color else v
-    d = d or kwargs or dict()
-    pairs = (f'{k}{sep}{_log_val(v)}' for k, v in d.items())
-    pref = log_s('{', c='m') if with_color else '{'
-    post = log_s('}', c='m') if with_color else '}'
-    return pref + ', '.join(pairs) + post
+                if is_float(v):  # Pad only normal, expected floats, intended for metric logging
+                    if is_float(v, no_int=True, no_sci=True):
+                        v = float(v)
+                        if with_color:
+                            return PrettyLogger.log(v, c='i', as_str=True, pad=pad_float)
+                        else:
+                            return f'{v:>{pad_float}}'
+                    else:
+                        return PrettyLogger.i(v) if with_color else v
+                else:
+                    return PrettyLogger.i(v) if with_color else v
+        d = d or kwargs or dict()
+        if for_path:
+            assert not with_color  # sanity check
+            sep = '='
+        if with_color:
+            sep = PrettyLogger.s(sep, c='m')
+        pairs = (f'{k}{sep}{_log_val(v)}' for k, v in d.items())
+        pref, post = '{', '}'
+        if with_color:
+            pref, post = PrettyLogger.s(pref, c='m'), PrettyLogger.s(post, c='m')
+        return pref + ', '.join(pairs) + post
 
 
-def log_dict_nc(d: Dict = None, **kwargs) -> str:
-    """
-    Syntactic sugar for no color
-    """
-    return log_dict(d, with_color=False, **kwargs)
-
-
-def log_dict_id(d: Dict) -> str:
-    """
-    Indented dict
-    """
-    return json.dumps(d, indent=4)
-
-
-def log_dict_pg(d: Dict) -> str:
-    """
-    prettier dict colored by `pygments` and with indent
-    """
-    return highlight(log_dict_id(d), lexers.JsonLexer(), formatters.TerminalFormatter())
-
-
-def log_dict_p(d: Dict, **kwargs) -> str:
-    """
-    a compact, one-line str for dict
-
-    Intended as part of filename
-    """
-    return log_dict(d, with_color=False, sep='=', **kwargs)
+pl = PrettyLogger()
 
 
 def hex2rgb(hx: str, normalize=False) -> Union[Tuple[int], Tuple[float]]:
@@ -352,12 +388,15 @@ def get_logger(name: str, typ: str = 'stdout', file_path: str = None) -> logging
     logger.setLevel(logging.DEBUG)
     if typ == 'stdout':
         handler = logging.StreamHandler(stream=sys.stdout)  # stdout for my own coloring
-    else:
+    else:  # `file-write`
+        if not file_path:
+            raise ValueError(f'{pl.i(file_path)} must be specified for {pl.i("file-write")} logging')
         os.makedirs(os.path.dirname(file_path), exist_ok=True)
         handler = logging.FileHandler(file_path)
     handler.setLevel(logging.DEBUG)
     handler.setFormatter(MyFormatter(with_color=typ == 'stdout'))
     logger.addHandler(handler)
+    logger.propagate = False
     return logger
 
 
@@ -365,7 +404,12 @@ class MlPrettier:
     """
     My utilities for deep learning training logging
     """
-    def __init__(self, ref: Dict[str, Any] = None, metric_keys: List[str] = None):
+    no_prefix = ('epoch', 'step')
+
+    def __init__(
+            self, ref: Dict[str, Any] = None, metric_keys: List[str] = None, no_prefix: Iterable[str] = no_prefix,
+            with_color: bool = True
+    ):
         """
         :param ref: Reference that are potentially needed
             i.e. for logging epoch/step, need the total #
@@ -374,6 +418,8 @@ class MlPrettier:
         """
         self.ref = ref
         self.metric_keys = metric_keys or ['acc', 'recall', 'auc']
+        self.no_prefix = no_prefix
+        self.with_color = with_color
 
     def __call__(self, d: Union[str, Dict], val=None) -> Union[Any, Dict[str, Any]]:
         """
@@ -385,11 +431,11 @@ class MlPrettier:
             raise ValueError('Either a key-value pair or a mapping is expected')
         if is_dict:
             d: Dict
-            return {k: self._pretty_single(k, v) for k, v in d.items()}
+            return {k: self.single(k, v) for k, v in d.items()}
         else:
-            return self._pretty_single(d, val)
+            return self.single(d, val)
 
-    def _pretty_single(self, key: str, val=None) -> Union[str, List[str], Dict[str, Any]]:
+    def single(self, key: str = None, val: Any = None) -> Union[str, List[str], Dict[str, Any]]:
         """
         `val` processing is infered based on key
         """
@@ -403,7 +449,10 @@ class MlPrettier:
             else:
                 fmt = f'%{len_lim + 4}.3f'
                 s_val = fmt % val
-            return f'{s_val}/{lim}'  # Pad integer
+            if self.with_color:
+                return f'{pl.i(s_val)}/{pl.i(lim)}'
+            else:
+                return f'{s_val}/{lim}'  # Pad integer
         elif 'loss' in key:
             return f'{round(val, 4):7.4f}'
         elif any(k in key for k in self.metric_keys):  # custom in-key-ratio metric
@@ -421,6 +470,18 @@ class MlPrettier:
         else:
             return val
 
+    def should_add_split_prefix(self, key: str) -> bool:
+        """
+        Whether to add split prefix to the key
+        """
+        return key not in self.no_prefix
+
+    def add_split_prefix(self, d: Dict[str, Any], split: str = None):
+        if split is None:
+            return d
+        else:
+            return {f'{split}/{k}' if self.should_add_split_prefix(k) else k: v for k, v in d.items()}
+
 
 class MyProgressCallback(TrainerCallback):
     """
@@ -448,8 +509,12 @@ class MyProgressCallback(TrainerCallback):
         return state.max_steps // state.num_train_epochs
 
     @staticmethod
-    def _get_curr_epoch(state) -> str:
-        return MlPrettier(ref=dict(epoch=state.num_train_epochs))('epoch', int(state.epoch+1))
+    def _get_curr_epoch(state, is_eval: bool = False) -> str:
+        n_ep = int(state.epoch)
+        if not is_eval:  # heuristic judging by the eval #epoch shown
+            n_ep += 1
+
+        return MlPrettier(ref=dict(epoch=state.num_train_epochs))('epoch', n_ep)
 
     def on_epoch_begin(self, args, state, control, **kwargs):
         if state.is_local_process_zero:
@@ -475,7 +540,7 @@ class MyProgressCallback(TrainerCallback):
         if not self.train_only:
             if state.is_local_process_zero and isinstance(eval_dataloader.dataset, Sized):
                 if self.prediction_bar is None:
-                    ep = MyProgressCallback._get_curr_epoch(state)
+                    ep = MyProgressCallback._get_curr_epoch(state, is_eval=True)
                     desc = f'Eval Epoch {ep}'
                     self.prediction_bar = tqdm(
                         desc=desc, total=len(eval_dataloader), leave=self.training_bar is None, unit='ba'
@@ -497,6 +562,123 @@ class MyProgressCallback(TrainerCallback):
         pass
 
 
+class CheckArg:
+    """
+    An easy, readable interface for checking string arguments as effectively enums
+
+    Intended for high-level arguments instead of actual data processing as not as efficient
+
+    Raise errors when common arguments don't match the expected values
+    """
+    logger = get_logger('Arg Checker')
+
+    def check_mismatch(self, display_name: str, val: str, accepted_values: List[str]):
+        if self.verbose:
+            d_log = dict(val=val, accepted_values=accepted_values)
+            CheckArg.logger.info(f'Checking {pl.i(display_name)} w/ {pl.i(d_log)}... ')
+        if val not in accepted_values:
+            raise ValueError(f'Unexpected {pl.i(display_name)}: '
+                             f'expect one of {pl.i(accepted_values)}, got {pl.i(val)}')
+
+    def __init__(self, ignore_none: bool = True, verbose: bool = False):
+        """
+        :param ignore_none: If true, arguments passed in as `None` will not raise error
+        :param verbose: If true, logging are print to console
+        """
+        self.d_name2func = dict()
+        self.ignore_none = ignore_none
+        self.verbose = verbose
+
+    def __call__(self, **kwargs):
+        for k, v in kwargs.items():
+            if self.ignore_none and v is None:
+                if self.verbose:
+                    CheckArg.logger.warning(f'Argument {pl.i(k)} is {pl.i("None")} and ignored')
+                continue
+            self.d_name2func[k](v)
+
+    def cache_mismatch(self, display_name: str, attr_name: str, accepted_values: List[str]):
+        self.d_name2func[attr_name] = lambda x: self.check_mismatch(display_name, x, accepted_values)
+
+
+ca = CheckArg()
+ca.cache_mismatch(  # See `stefutil::plot.py`
+    'Bar Plot Orientation', attr_name='bar_orient', accepted_values=['v', 'h', 'vertical', 'horizontal']
+)
+
+
+def now(as_str=True, for_path=False, fmt: str = 'full') -> Union[datetime.datetime, str]:
+    """
+    # Considering file output path
+    :param as_str: If true, returns string; otherwise, returns datetime object
+    :param for_path: If true, the string returned is formatted as intended for file system path
+        relevant only when as_str is True
+    :param fmt: One of [`full`, `date`, `short-date`]
+        relevant only when as_str is True
+    """
+    d = datetime.datetime.now()
+    if as_str:
+        ca.check_mismatch('Date Format', fmt, ['full', 'date', 'short-date'])
+        if fmt == 'full':
+            fmt_tm = '%Y-%m-%d_%H-%M-%S' if for_path else '%Y-%m-%d %H:%M:%S.%f'
+        else:
+            fmt_tm = '%Y-%m-%d'
+        ret = d.strftime(fmt_tm)
+        if fmt == 'short-date':  # year in 2-digits
+            ret = ret[2:]
+        return ret
+    else:
+        return d
+
+
 if __name__ == '__main__':
-    lg = get_logger('test')
-    lg.info('test')
+    # lg = get_logger('test')
+    # lg.info('test')
+
+    def check_log_lst():
+        lst = ['sda', 'asd']
+        print(pl.i(lst))
+        # with open('test-logi.txt', 'w') as f:
+        #     f.write(pl.nc(lst))
+    check_log_lst()
+
+    def check_log_tup():
+        tup = ('sda', 'asd')
+        print(pl.i(tup))
+    check_log_tup()
+
+    def check_logi():
+        d = dict(a=1, b=2)
+        print(pl.i(d))
+    check_logi()
+
+    def check_nested_log_dict():
+        d = dict(a=1, b=2, c=dict(d=3, e=4, f=['as', 'as']))
+        mic(d)
+        print(pl.i(d))
+        print(pl.nc(d))
+        mic(pl.i(d), pl.nc(d))
+    check_nested_log_dict()
+
+    def check_logger():
+        logger = get_logger('blah')
+        logger.info('should appear once')
+    # check_logger()
+
+    def check_now():
+        mic(now(fmt='full'))
+        mic(now(fmt='date'))
+        mic(now(fmt='short-date'))
+    # check_now()
+
+    def check_ca():
+        ori = 'v'
+        ca(bar_orient=ori)
+    # check_ca()
+
+    def check_ca_warn():
+        ca_ = CheckArg(verbose=True)
+        ca_.cache_mismatch(display_name='Disp Test', attr_name='test', accepted_values=['a', 'b'])
+        ca_(test='a')
+        ca_(test=None)
+    check_ca_warn()
